import re
from typing import Dict, List

from lxml import etree

from ..common import Element, TcTool


class XmlSorter(TcTool):
    """Tool to sort XML files.

    Use one instance for a sequence of files.
    """

    LOGGER_NAME = "xml_sorter"

    FILTER_DEFAULT: List[str] = ["*.tsproj", "*.xti", "*.plcproj"]

    CONFIG_KEY = "xml_sort"

    def __init__(self, *args):
        super().__init__(*args)

        if self.args.skip_nodes is None:
            self.args.skip_nodes = []

        self._file_changed = False  # True if any change is made in the current path
        # This is a property to avoid passing around booleans between recursive calls

    @classmethod
    def set_arguments(cls, parser):
        super().set_arguments(parser)

        parser.prog = "tc_xml_sort"
        parser.description = "Alphabetically sort the nodes in XML files."
        parser.epilog = "Example: ``tc_xml_sort -r ./MyTwinCATProject``"
        parser.add_argument(
            "--skip-nodes",
            "-n",
            help="Do not touch the attributes and sub-nodes of nodes with these names",
            nargs="+",
            default=["Device", "DataType", "DeploymentEvents"],
        )
        return parser

    def run(self) -> int:
        for file in self.find_files():
            self.sort_file(str(file))

        self.logger.info(f"Checked {self.files_checked} path(s)")

        if self.args.check:
            if self.files_to_alter == 0:
                self.logger.info("No changes to be made in checked files!")
                return 0

            self.logger.info(f"{self.files_to_alter} path(s) can be re-sorted")
            return 1

        self.logger.info(f"Re-saved {self.files_resaved} path(s)")

    def sort_file(self, path: str):
        """Sort a single path."""
        tree = self.get_xml_tree(path)

        self.files_checked += 1
        self._file_changed = False  # Reset

        root = tree.getroot()
        self.sort_node_recursively(root)

        # Re-indent by a double space
        etree.indent(tree, space="  ", level=0)

        self.logger.debug(f"Processing path `{path}`...")

        tree_bytes = etree.tostring(root, doctype=self.header_before)

        with open(path, "rb") as fh:
            current_contents = fh.readlines()

        current_bytes = b"".join(current_contents)

        if self._file_changed:
            self.files_to_alter += 1

        if current_bytes != tree_bytes:
            if self.args.dry:
                self.logger.debug(f"Old path contents of `{path}`:")
                self.logger.debug("-" * 50)
                self.logger.debug(current_bytes.decode("utf-8"))
                self.logger.debug("-" * 50)
                self.logger.debug("New path contents:")
                self.logger.debug("-" * 50)
                self.logger.debug(tree_bytes.decode("utf-8"))
                self.logger.debug("-" * 50)

            self.logger.debug(f"File can be re-sorted: `{path}`")

            if not self.args.check and not self.args.dry:
                with open(path, "wb") as fh:
                    fh.write(tree_bytes)
                    # Write by hand (instead of `tree.write()` so we control the header
                    self.files_resaved += 1
        else:
            if self.args.dry:
                self.logger.debug("Content identical for `{path}`")

    def sort_node_recursively(self, node: Element):
        """Sort a node and any sub-nodes, and their sub-nodes.

        Sorting is done in-place, the object is passed in by reference.
        """

        if self.get_tag(node) in self.args.skip_nodes:
            return  # Stop here

        if self.get_attrib(node).get("space", None):
            return  # Do not touch with `xml:space="preserve"`

        # Also sort the attributes - but this won't work flawlessly, since dicts are
        # inherently unsorted
        if self.sort_attributes(node):
            self._file_changed = True

        # First sort the next level of children, necessary to sort the current node too
        for child in node:
            self.sort_node_recursively(child)

        new_children = sorted(node, key=self.get_node_sorting_key)

        if new_children != node[:]:
            self._file_changed = True

        node[:] = new_children  # Replace children in place

    @staticmethod
    def sort_attributes(node: Element) -> bool:
        """Sort the attributes of a node.

        :returns: True if any changes were really made
        """
        sorted_attrs = sorted(node.attrib.items())
        changed = sorted_attrs != node.attrib.items()
        node.attrib.clear()
        node.attrib.update(sorted_attrs)
        return changed

    @staticmethod
    def get_node_sorting_key(node: Element) -> str:
        """Get the string by which sub-nodes will be sorted.

        Sorting will be done on the literal node XML subtree string.
        """
        key = etree.tostring(node, encoding="unicode")
        key = re.sub(r"\s+", "", key, flags=re.UNICODE)

        return key

    @staticmethod
    def get_tag(node: Element) -> str:
        """Get tag without URL prefix from node."""
        tag = node.tag
        if tag.startswith("{"):
            # Keep only the part after the `{...}`
            _, _, tag = tag.partition("}")

        return tag

    @staticmethod
    def get_attrib(node: Element) -> Dict[str, str]:
        """Yield node attributes, with namespace stripped."""
        attributes = {}
        for key, value in node.attrib.items():
            if key.startswith("{"):
                # Keep only the part after the `{...}`
                _, _, key = key.partition("}")
            attributes[key] = value

        return attributes



import re
from collections import OrderedDict
from typing import List, Optional, Tuple, Type

from editorconfig import get_properties

from ..common import TcTool
from .format_extras import Kind
from .format_rules import (
    FormatConditionalParentheses,
    FormatEndOfLine,
    FormatInsertFinalNewline,
    FormatTabs,
    FormattingRule,
    FormatTrailingWhitespace,
    FormatVariablesAlign,
)

RowCol = Tuple[int, int]
Segment = Tuple[Kind, List[str], str]


class XmlMachine:
    """Helper class to identify code bits inside an XML path."""

    _re_name = re.compile(r'Name="(\w+)"')

    def __init__(self):
        self._kind = Kind.XML
        self._name = ""
        self._row = 0  # Line number inside path
        self._col = 0  # Position inside line

        self.regions: List[Tuple[RowCol, Kind, str]] = []

    def parse(self, content: List[str]):
        """Progress machine line by line."""
        self._kind = Kind.XML
        self._row = 0
        self._col = 0
        self.regions = [((self._row, self._col), Kind.XML, "<unknown>")]
        for self._row, line in enumerate(content):
            self._col = 0

            if matches := self._re_name.search(line):
                self._name = matches.group(1)

            while self._col < len(line):
                self._parse_line(line)

    def _parse_line(self, line: str):
        """Parse a line, not necessarily starting from the left.

        :param line:
        """
        pos_before = self._col
        if self._kind == Kind.XML:
            self._find_state_in_line(line, "<Declaration><![CDATA[", Kind.DECLARATION)
            self._find_state_in_line(line, "<ST><![CDATA[", Kind.IMPLEMENTATION)
        elif self._kind == Kind.DECLARATION:
            self._find_state_in_line(line, "]]></Declaration>", Kind.XML, True)
        elif self._kind == Kind.IMPLEMENTATION:
            self._find_state_in_line(line, "]]></ST>", Kind.XML, True)

        if self._col == pos_before:
            self._col = len(line)  # Nothing found, continue to next line

    def _find_state_in_line(
        self, line, key: str, new_state: Kind, include_key: bool = False
    ):
        """Find key elements inside a line to advance the states.

        When ``include_key`` is True, the text of the key is considered part of the new
        state.
        """
        pos = line.find(key, self._col)
        if pos >= 0:
            self._kind = new_state
            self._col = pos + len(key)

            if not include_key:
                pos += len(key)

            self.regions.append(((self._row, pos), self._kind, self._name))
            # First character of the new region


class Formatter(TcTool):
    """Helper to check formatting in PLC files.

    Instantiate once for a sequence of files.
    """

    LOGGER_NAME = "formatter"

    FILTER_DEFAULT = ["*.TcPOU", "*.TcGVL", "*.TcDUT"]

    CONFIG_KEY = "format"

    _RULE_CLASSES: List[Type[FormattingRule]] = []

    def __init__(self, *args):
        super().__init__(*args)

        # Keep some dynamic properties around just so we don't have to constantly pass
        # them between methods
        self._file = ""
        self._properties = OrderedDict()
        self._rules: List[FormattingRule] = []

        self._number_corrections = 0  # Track number of changes for the current file

    @classmethod
    def set_arguments(cls, parser):
        super().set_arguments(parser)

        parser.description = "Format the PLC code inside a TwinCAT source XML path."
        parser.epilog = "Example: ``tc_format -r ./MyTwinCATProject``"
        return parser

    @classmethod
    def register_rule(cls, new_rule: Type[FormattingRule]):
        """Incorporate a new formatting rule (accounting for its priority)."""
        cls._RULE_CLASSES.append(new_rule)
        sorted(cls._RULE_CLASSES, key=lambda item: item.PRIORITY)

    def run(self) -> int:
        files = self.find_files()

        for file in files:
            self.format_file(str(file))

        self.logger.info(f"Checked {self.files_checked} path(s)")

        if self.args.check:
            if self.files_to_alter == 0:
                self.logger.info("No changes to be made in checked files!")
                return 0

            self.logger.info(f"{self.files_to_alter} path(s) should altered")
            return 1

        self.logger.info(f"Re-saved {self.files_resaved} path(s)")
        return 0

    def format_file(self, path: str):
        """Format (or check) a specific path.

        The path is read as text and code inside XML tags is detected manually. Other
        lines of XML remain untouched.
        """
        with open(path, "r", encoding="utf-8", errors="ignore", newline="") as fh:
            content = fh.readlines()

        self._file = path

        self._properties = get_properties(path)
        if not self._properties:
            self.logger.warning(f"Editorconfig properties is empty for file `{path}`")

        self.files_checked += 1
        self._number_corrections = 0

        self.logger.debug(f"Processing path `{path}`...")

        self._rules = [rule(self._properties) for rule in self._RULE_CLASSES]

        # Do whole-file rules first:
        for rule in self._rules:
            if rule.WHOLE_FILE:
                self.apply_rule(rule, content)

        segments: List[Segment] = list(self.split_code_segments(content))

        for kind, segment, _ in segments:
            # Changes are done in-place
            self.format_segment(segment, kind)

        if self._number_corrections > 0:
            self.files_to_alter += 1

        if not self.args.dry and not self.args.check and self._number_corrections > 0:
            with open(path, "w", newline="", encoding="utf-8") as fh:
                # Keep newline symbols inside strings
                for _, segment, _ in segments:
                    fh.write("".join(segment))

            self.files_resaved += 1

    @staticmethod
    def split_code_segments(content: List[str]):
        """Copy content, split into XML and code sections.

        Function is a generator, each pair is yielded.

        Note: line endings are not modified! I.e., segments should be appended together
        directly, without extra newlines.

        :param: File content as list
        :return: List[Segment]
        """
        if not content:
            return  # Nothing to yield

        machine = XmlMachine()
        machine.parse(content)

        regions = machine.regions
        regions.append(
            ((len(content), len(content[-1])), Kind.XML, "<unknown>")
        )  # Add end-of-path

        # Iterate over pairs of regions so we got the start and end together
        for (rowcol_prev, kind_prev, name_prev), (rowcol, _, _) in zip(
            regions[:-1], regions[1:]
        ):
            lines = content[rowcol_prev[0] : (rowcol[0] + 1)]  # Inclusive range

            if rowcol[0] == rowcol_prev[0]:  # If there is only a single line!
                lines[0] = lines[0][rowcol_prev[1] : rowcol[1]]
                # Use two columns in one step, otherwise the columns move
            else:
                if rowcol_prev[1] > 0:
                    lines[0] = lines[0][rowcol_prev[1] :]
                    # Keep end of the first line

                if rowcol[1] < len(lines[-1]):
                    lines[-1] = lines[-1][: rowcol[1]]
                    # Keep start of the last line (last character is not included!)

            yield kind_prev, lines, name_prev

    def format_segment(self, content: List[str], kind: Kind):
        """Format a specific segment of code.

        :param content: Text to reformat (changed in place!)
        :param kind: Type of the content
        """
        if kind == Kind.XML:
            return  # Do nothing
        else:
            for rule in self._rules:
                if not rule.WHOLE_FILE:  # Skip otherwise
                    self.apply_rule(rule, content, kind)

    def apply_rule(self, rule, content, kind: Optional[Kind] = None):
        """Run a rule over some content and handle results."""
        rule.format(content, kind)
        corrections = rule.consume_corrections()
        self._number_corrections += len(corrections)

        tag = f"[{kind.name.lower()}]" if kind else "[file]"

        for line_nr, message in corrections:
            # `line_r` is zero-indexed
            self.logger.debug(f"{self._file}{tag}:{line_nr + 1}\t{message}")


Formatter.register_rule(FormatTabs)
Formatter.register_rule(FormatTrailingWhitespace)
Formatter.register_rule(FormatInsertFinalNewline)
Formatter.register_rule(FormatEndOfLine)
Formatter.register_rule(FormatVariablesAlign)
Formatter.register_rule(FormatConditionalParentheses)



import re
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List, Optional

from git import Repo
from lxml import etree

from ..common import Tool

ElementTree = etree._ElementTree  # noqa
Element = etree._Element  # noqa


class MakeRelease(Tool):
    """Tool to create a release archive from a TwinCAT project."""

    LOGGER_NAME = "make_release"

    CONFIG_KEY = "make_release"

    def __init__(self, *args):
        super().__init__(*args)

        # Bunch of attributes to easily share data between methods:
        self.version: Optional[str] = None
        self.destination_dir: Optional[Path] = None
        self.archive_source: Optional[Path] = None
        self.config_dir: Optional[Path] = None

    @classmethod
    def set_arguments(cls, parser):
        super().set_arguments(parser)
        parser.description = """Create a release archive from the current project.

Archives the compiled result of PlC and HMI projects.
Checks can be performed to make sure the binaries work on a target PLC.

The resulting archive will be named after the PLC project.
"""

        parser.add_argument(
            "plc_source",
            metavar="plc-source",
            help="Path where to search for compilation directory of PLC project",
        )

        parser.add_argument(
            "--destination",
            help="Folder where release output will be saved (default: ./deploy)",
            default="deploy",
        )

        parser.add_argument(
            "--include-hmi",
            help="Also include HMI compilation in release (default: False)",
            default=False,
            action="store_true",
        )

        parser.add_argument(
            "--add-file",
            "-a",
            help="Add additional file to release package "
            "(<filepath> <relative path inside archive>)",
            default=None,
            nargs="+",  # Can be 1 or 2
            action="append",
        )

        parser.add_argument(
            "--platform",
            help="Target platform for PLC to copy (default: `x64`)",
            default="x64",
        )

        parser.add_argument(
            "--check-cpu",
            help="Validate the CPU configuration in the compiled project "
            "(<number of cores> <number of isolated cores>)",
            nargs=2,
            type=int,
            default=None,
        )

        parser.add_argument(
            "--check-devices",
            help="Validate devices, only the listed devices by name may be enabled",
            nargs="+",
            default=None,
        )

        parser.add_argument(
            "--check-version-variable",
            help="Validate current version in PLC source code in the given variable "
            "(<filename> <variable name>)",
            nargs=2,
            default=None,
        )

        parser.add_argument(
            "--check-version-hmi",
            help="Validate current version in HMI source code in the given object "
            "(<filename> <object id>)",
            nargs=2,
            default=None,
        )

    @staticmethod
    def glob_first(path: Path, pattern: str) -> Path:
        """Get the first glob result."""
        results = path.rglob(pattern)
        return next(results)

    def run(self) -> int:
        """Create release archive."""
        source_dir = Path(self.args.plc_source)

        repo = Repo(source_dir, search_parent_directories=True)
        repo_dir = Path(repo.git_dir).parent

        self.logger.debug(f"Found Git repository `{repo_dir}`")

        self.version = repo.git.tag()
        if not self.version:
            self.logger.warning(f"Could not find any tags in `{repo}`")
            self.version = "v0.0.0"

        self.logger.info(f"Making release for tag `{self.version}`")

        self.destination_dir = Path(self.args.destination).absolute()

        if not self.destination_dir.is_dir():
            self.destination_dir.mkdir(parents=True)

        self.logger.debug(f"Going to make release in `{self.destination_dir}`")

        pattern = f"_Boot/*({self.args.platform})"
        boot_dir = self.glob_first(source_dir, pattern)
        self.logger.debug(f"Copying PLC boot files from `{boot_dir}")

        plc_project = self.glob_first(boot_dir, "*.tpzip")
        name = plc_project.stem.lower().replace(" ", "_")

        hmi_bin_dir: Optional[Path] = None
        if self.args.include_hmi:
            html_file = self.glob_first(source_dir, "bin/*.html")
            hmi_bin_dir = html_file.parent

        archive_file = self.destination_dir / f"{name}-{self.version}.zip"
        if archive_file.is_file():
            raise RuntimeError(f"Target file `{archive_file}` already exists")

        # Create self-deleting temporary folder inside the release directory:
        with TemporaryDirectory(dir=self.destination_dir.parent) as temp_dir_str:
            temp_dir = Path(temp_dir_str)

            # Copy entire boot directory content to new folder
            self.archive_source = temp_dir / "release"
            shutil.copytree(boot_dir, self.archive_source / "PLC", dirs_exist_ok=True)

            # Also copy HMI bin directory:
            if hmi_bin_dir:
                shutil.copytree(
                    hmi_bin_dir,
                    self.archive_source / "HMI",
                    dirs_exist_ok=True,
                )

            errors = self.validate_release(temp_dir)
            for error in errors:
                self.logger.error(error)

            self.add_additional_files()

            if self.args.dry:
                return 0  # Don't make any more changes

            if errors:
                self.logger.error(
                    f"Not making release because of {len(errors)} failed check(s)"
                )
                return 1

            # `make_archive` tends to add itself too, so create it outside the source:
            shutil.make_archive(
                str(archive_file.with_suffix("")), "zip", self.archive_source
            )

        self.logger.info(f"Created file `{archive_file}`")
        return 0

    def add_additional_files(self):
        if not self.args.add_file:
            return

        for file_option in self.args.add_file:
            file_source = Path(file_option[0])
            if len(file_option) == 1:
                file_dest = file_source
            elif len(file_option) == 2:
                file_dest = Path(file_option[1])
            else:
                raise RuntimeError("`--add-file` argument must have 1 or 2 values")

            if file_dest.is_absolute():
                file_dest = file_dest.relative_to(Path.cwd())

            shutil.copy(file_source, self.archive_source / file_dest)

        return

    def validate_release(self, temp_dir: Path) -> List[str]:
        """

        :param temp_dir: Root of temporary directory
        """
        # Unpack CurrentConfig into temp dir:
        self.config_dir = temp_dir / "CurrentConfig"
        shutil.unpack_archive(
            self.archive_source / "PLC" / "CurrentConfig.tszip",
            self.config_dir,
            format="zip",
        )

        project_file = self.glob_first(self.config_dir, "*.tsproj")
        root = etree.parse(project_file)

        errors = []
        errors += self.check_cpu(root)
        errors += self.check_devices(root)
        errors += self.check_version_variable(temp_dir)
        errors += self.check_version_hmi()

        return errors

    def check_cpu(self, root: ElementTree) -> List[str]:
        """Validate CPU configuration."""
        if self.args.check_cpu is None:
            return []

        node: Element = root.xpath("//TcSmProject/Project/System/Settings")[0]
        cpus = [
            int(node.attrib[key]) if key in node.attrib else 0
            for key in ["MaxCpus", "NonWinCpus"]
        ]
        expected_cpus = self.args.check_cpu

        if cpus[0] != expected_cpus[0] or cpus[1] != expected_cpus[1]:
            return [
                f"Expected cpu configuration {expected_cpus}, but found {cpus} in "
                f"project file",
            ]

        return []

    def check_devices(self, root: ElementTree) -> List[str]:
        """Validate device configuration."""
        errors = []

        if self.args.check_devices is None:
            return errors

        devices: List[Element] = root.xpath("//TcSmProject/Project/Io/Device")

        for i, device in enumerate(devices):
            if "File" in device.attrib:  # Replace by file reference
                extra_file = self.glob_first(self.config_dir, device.attrib["File"])
                extra_root: ElementTree = etree.parse(extra_file)
                new_device: Element = extra_root.xpath("//TcSmItem/Device")[0]
                new_device.find("Name").text = device.attrib["File"].rstrip(".xti")
                devices[i] = new_device

        for device in devices:
            name = device.find("Name").text
            should_be_disabled = name not in self.args.check_devices
            is_disabled = device.attrib.get("Disabled", "false").lower() == "true"
            if should_be_disabled != is_disabled:
                msg = (
                    "disabled, but is enabled"
                    if should_be_disabled
                    else "enabled, but is disabled"
                )
                errors.append(f"Device `{name}` should be {msg}!")

        return errors

    def check_version_variable(self, temp_dir: Path) -> List[str]:
        """Validate the version variable matches the release version.

        :param temp_dir:
        """
        if self.args.check_version_variable is None:
            return []

        check_file, check_variable = self.args.check_version_variable

        plc_archive = self.glob_first(
            self.archive_source / "PLC" / "CurrentConfig",
            "*.tpzip",
        )
        # Unpack CurrentConfig/<plc>.tpzip into temp dir:
        plc_dir = self.config_dir / "plc"
        shutil.unpack_archive(plc_archive, plc_dir, format="zip")

        file = self.glob_first(plc_dir, check_file)
        pattern = re.compile(check_variable + r".*" + self.version.replace(".", r"\."))

        contents = file.read_text()
        matches = pattern.findall(contents)

        if not matches:
            return [
                f"Failed to find version `{self.version}` in code "
                f"`{check_file}:{check_variable}`"
            ]

        return []

    def check_version_hmi(self):
        """Validate version string inside HMI page."""
        if self.args.check_version_hmi is None:
            return []

        check_file, check_object = self.args.check_version_hmi

        file = self.glob_first(self.archive_source / "HMI", check_file)
        pattern = re.compile(check_object + r'.*data-tchmi-text="(.+?)"')
        # Use `+?` for lazy matching

        contents = file.read_text()
        for match in pattern.finditer(contents):
            if match.group(1) == self.version:
                return []
            return [
                f"Version in HMI `{check_file}:{check_object}` is "
                f"`{match.group(1)}`, not `{self.version}`"
            ]

        return [f"Failed to find HMI object `{check_object}` in `{check_file}`"]



import sys

from .format_class import Formatter


def main(*args) -> int:
    formatter = Formatter(*args)
    return formatter.run()


def main_argv():
    """Entrypoint for the executable, defined through ``pyproject.toml``."""
    exit(main(*sys.argv[1:]))


def get_parser():
    return Formatter.get_argument_parser()


if __name__ == "__main__":
    exit_code = main(*sys.argv[1:])  # Skip script name
    exit(exit_code)



import argparse

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="tctools",
        description="Use either any of [`tctools.format`, `tctools.git_info`, "
        "`tctools.make_release`, `tctools.xml_sorter`]",
    )

    parser.print_help()



import math
import re
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, OrderedDict, Tuple, Type

from .format_extras import Kind

Correction = Tuple[int, str]


class FormattingRule(ABC):
    """TcFormatter rule base class.

    Extend and implement this class to check for and correct a specific error/style/etc.

    :cvar PRIORITY: Lower priority means a rule gets applied earlier.
    :cvar WHOLE_FILE: If True, rule is applied to the entire file instead of just
                      code blocks.
    """

    PRIORITY = 100
    WHOLE_FILE = False

    def __init__(self, properties: OrderedDict):
        self._properties = properties
        self._corrections: List[Correction] = []

        # Universal properties:

        # Number of spaces per indentation:
        self._indent_size: int = self.get_property("indent_size", 4, value_type=int)
        # Number of spaces per tab character:
        self._tab_width: int = self.get_property(
            "tab_width", default=self._indent_size, value_type=int
        )

        self._indent_style: Optional[str] = self._properties.get("indent_style", None)

        self._indent_str: str = " " * self._indent_size
        if self._indent_style and self._indent_style == "tab":
            self._indent_str = "\t"

        self._end_of_line: Optional[str] = self._properties.get("end_of_line", None)
        options = {"lf": "\n", "cr": "\r", "crlf": "\r\n"}
        self._line_ending: str = options.get(self._end_of_line, "\n")

        self._re_any_line_ending = re.compile(r"(\r\n|\n|\r)")  # Find any full EOL

    @property
    def actual_indent_size(self) -> int:
        """Tab width for style=tabs or indent size for style=spaces."""
        if self._indent_style == "tab":
            return self._tab_width

        return self._indent_size

    def get_property(
        self,
        name: str,
        default: Any = None,
        value_type: Optional[Type] = None,
    ) -> Any:
        """Get item from ``_properties``, parsing as needed.

        :param name:
        :param default: Value to dfault if name doesn't exist
        :param value_type: Class of the returned value (e.g. ``bool``)
        """
        if name not in self._properties:
            return default

        value = self._properties[name]
        if value_type is None:
            return value  # Unprocessed

        if value_type == bool:
            if isinstance(value, str):
                return value in ["TRUE", "True", "true", "1"]
            return bool(value)

        return value_type(value)

    @abstractmethod
    def format(self, content: List[str], kind: Optional[Kind] = None):
        """Fun rule to format text.

        :param content: Text to format (changed in place!)
        :param kind:    Kind of content
        """
        pass

    def add_correction(self, message: str, line_nr: int):
        """Register a formatting correction.

        See :class:`Correction`.
        """
        self._corrections.append((line_nr, message))

    def consume_corrections(self) -> List[Correction]:
        """Return listed corrections and reset list."""
        corrections = self._corrections
        self._corrections = []
        return corrections


class FormatTabs(FormattingRule):
    """Check usage of tab character."""

    _re_tab = re.compile(r"\t")
    _re_spaces = re.compile(r"  +")  # Match two spaces or more

    def __init__(self, *args):
        super().__init__(*args)

    def format(self, content: List[str], kind: Optional[Kind] = None):
        if self._indent_style == "tab":
            re_search = self._re_spaces
        elif self._indent_style == "space":
            re_search = self._re_tab
        else:
            return

        for i, line in enumerate(content):
            pos = 0
            count = 0
            while (matches := re_search.search(line, pos)) is not None:
                pos = matches.end()
                num_chars = (
                    int(math.ceil((matches.end() - matches.start()) / self._tab_width))
                    if self._indent_str == "\t"
                    else self._indent_size - (matches.start() % self._indent_size)
                )

                line = (
                    line[: matches.start()]
                    + self._indent_str[0] * num_chars
                    + line[matches.end() :]
                )
                count += 1

            if count > 0:
                self.add_correction(
                    (
                        "Line contains an indent that should be a tab"
                        if self._indent_str == "\t"
                        else "Line contains a tab that should be spaces"
                    ),
                    i,
                )
                content[i] = line


class FormatTrailingWhitespace(FormattingRule):
    """Remove trailing whitespace."""

    PRIORITY = 90  # Precede `FinalNewline`

    _re_trailing_ws = re.compile(r"([^\S\r\n]+)([\r\n])*$")
    # Match whitespace (but not newlines) before (any) newline at the end of the line

    def __init__(self, *args):
        super().__init__(*args)

        self._remove_tr_ws = self.get_property(
            "trim_trailing_whitespace", False, value_type=bool
        )

    def format(self, content: List[str], kind: Optional[Kind] = None):
        if not self._remove_tr_ws:
            return  # Nothing to do
        for i, line in enumerate(content):
            line, count = re.subn(self._re_trailing_ws, r"\2", line)  # Keep group #2
            if count:
                content[i] = line
                self.add_correction("Line contains trailing whitespace", i)


class FormatInsertFinalNewline(FormattingRule):
    """Asserting a final empty newline in a file."""

    def __init__(self, *args):
        super().__init__(*args)

        self._insert_final_newline = self.get_property(
            "insert_final_newline", False, value_type=bool
        )

    def format(self, content: List[str], kind: Optional[Kind] = None):
        if not self._insert_final_newline:
            return

        if not content or len(content) == 1 and content[0] == "":
            return

        idx = len(content) - 1
        while idx >= 0:
            if content[idx].endswith("\n") or content[idx].endswith("\r"):
                return  # Newline found
            if content[idx] == "":
                idx -= 1  # Empty line, try the one before
            else:
                break  # Last content should be a newline, stay in function

        match = self._re_any_line_ending.search(content[0])
        # Get present EOL from first line
        eol = match.group(1) if match else self._line_ending

        content[-1] += eol
        self.add_correction("Block does not end with a newline", len(content) - 1)


class FormatEndOfLine(FormattingRule):
    """Asserting line endings are as expected."""

    WHOLE_FILE = True
    PRIORITY = 50  # Better do it a bit early

    def __init__(self, *args):
        super().__init__(*args)

        self._re_line_end = None

        if self._end_of_line is not None:
            if self._end_of_line == "lf":
                self._re_line_end = re.compile(r"\r\n|\r")
                # Works because Windows is first in the list
            elif self._end_of_line == "cr":
                self._re_line_end = re.compile(r"\r\n|\n")
            elif self._end_of_line == "crlf":
                self._re_line_end = re.compile(r"\r(?!\n)|(?<!\r)\n")
                # Match "\r" NOT followed by "\n" and match "\n" NOT preceded by "\r"
            else:
                raise ValueError(f"Unrecognized file ending `{self._line_ending}`")

    def format(self, content: List[str], kind: Optional[Kind] = None):
        if self._end_of_line is None:
            return  # Nothing specified

        count = 0
        for i, line in enumerate(content):
            line, new = re.subn(self._re_line_end, self._line_ending, line)
            if new > 0:
                content[i] = line
                count += new

        eol = self._line_ending.encode("unicode_escape")

        if count > 0:
            self.add_correction(
                f"{count} line endings need to be corrected to {eol}`", 0
            )


class FormatVariablesAlign(FormattingRule):
    """Assert whitespace align in variable declarations.

    Target formatting will create columns on the ":" and the "//" of comments.
    """

    PRIORITY = 110  # Go after FormatTabs

    def __init__(self, *args):
        super().__init__(*args)

        self._align = self.get_property(
            "twincat_align_variables", False, value_type=bool
        )

        self._re_variable = re.compile(
            r"""
                ^\s*                # Start of string + any ws
                (\S+)               # Sequence of non-ws
                \s*:                # Any ws + literal ":"
                \s*(.+?);           # Any ws + any sequence + literal ";"
                \s*([^\r\n]+)?      # Any ws + (Optional) any sequence
        """,
            re.VERBOSE,
        )

        self._re_newlines = re.compile(r"[\r\n]+$")

    def format(self, content: List[str], kind: Optional[Kind] = None):
        if not self._align:
            return  # Disabled by config

        if kind is None or kind is not Kind.DECLARATION:
            return  # Don't touch, only affect variable listing

        self.format_argument_list(content)

    def format_argument_list(self, content: List[str]):
        """Format entire declaration section"""

        # Get variable definitions, split up and keyed by content index:
        variable_definitions: Dict[int, List[Optional[str]]] = {}

        # Biggest size of each chunk across all lines:
        max_chunk_sizes: List[Optional[int]] = [None] * 3

        for i, line in enumerate(content):
            match = self._re_variable.match(line)
            if not match:
                continue

            chunks = list(match.groups())
            chunks[1] = f": {chunks[1]};"  # Bring match the matched characters

            variable_definitions[i] = chunks
            for j, chunk in enumerate(chunks):
                if not chunk:
                    continue
                if max_chunk_sizes[j] is None or len(chunk) > max_chunk_sizes[j]:
                    max_chunk_sizes[j] = len(chunk)

        if not variable_definitions:
            return  # No variables found, nothing to do

        new_indent = 1  # Variable name should start with one tab
        chunk_indent_levels = [new_indent]  # Number of indentations for each chunk
        for size in max_chunk_sizes[:-1]:
            new_indent += math.ceil((size + 2) / self.actual_indent_size)
            chunk_indent_levels.append(new_indent)

        for i, line_chunks in variable_definitions.items():
            new_line = ""
            for chunk, indent in zip(line_chunks, chunk_indent_levels):
                if chunk:
                    if indent > 0:
                        new_line += self._pad_to_indent_level(new_line, indent)
                    new_line += chunk

            if match_eol := self._re_newlines.search(content[i]):
                # The newline didn't get matched in the variable chunks, put it back:
                new_line += match_eol.group()

            if content[i] != new_line:
                self.add_correction("Variable declaration needs alignment", i)

            content[i] = new_line

    def _get_indent_string(self, col=0) -> str:
        """Return indent character(s), based on settings and current column.

        Either a tab or a set of spaces is returned such that the next tab index is
        reached exactly.
        """
        if self._indent_str == "\t":
            return self._indent_str

        num = self._indent_size - col % self._indent_size
        return self._indent_str[0] * num

    def _pad_to_indent_level(self, line: str, tab_index: int):
        """Add indents to the end of a string to reach a tab index.

        ``end_level = 1`` would pad until ``line`` is e.g. 4 characters.
        """
        index = math.floor(len(line) / self.actual_indent_size)

        padding_str = ""
        for i in range(tab_index - index):
            new_indent = (
                self._get_indent_string(col=len(line)) if i == 0 else self._indent_str
            )
            padding_str += new_indent

        return padding_str


class FormatConditionalParentheses(FormattingRule):
    """Formatter to make uses of parentheses inside IF, CASE and WHILE consistent.

    First regex is used to find potential corrections, which are then investigated
    by a loop to make sure parentheses remain matching and no essential parentheses
    are removed.
    """

    def __init__(self, *args):
        super().__init__(*args)

        self._parentheses = self.get_property(
            "twincat_parentheses_conditionals", value_type=bool
        )

        # Regex to find likely missing parentheses:
        self._re_needs_parentheses = re.compile(
            r"""
                # Look for start of string or new line:
                ^
                # Match keyword with surrounding ws:
                \s*(?:IF|WHILE|CASE)\s+
                # Match any characters NOT starting with "("
                # We cannot match the closing bracket, as this could be from a
                # function call
                ([^(\r\n].+?)
                # Match keyword with preceding ws:
                \s+(?:THEN|DO|OF)
            """,
            re.VERBOSE | re.MULTILINE,
        )

        # Regex to find likely redundant parentheses:
        self._re_removes_parentheses = re.compile(
            r"""
                # Look for start of string or new line:
                ^
                # Match keyword with surrounding ws:
                \s*(?:IF|WHILE|CASE)\s*
                # Match any characters within ():
                \((.+)\)
                # Match THEN with preceding ws:
                \s*(?:THEN|DO|OF)
            """,
            re.VERBOSE | re.MULTILINE,
        )

    def format(self, content: List[str], kind: Optional[Kind] = None):
        if self._parentheses is None:
            return  # Nothing to do

        pattern = (
            self._re_needs_parentheses
            if self._parentheses
            else self._re_removes_parentheses
        )

        for i, line in enumerate(content):
            # Do a manual match + replace, instead of e.g. subn(), because we might
            # need to add extra spaces after removing parentheses
            if match := pattern.search(line):
                prefix = line[: match.start(1)]
                condition = match.group(1)
                suffix = line[match.end(1) :]

                if self._parentheses:
                    condition = "(" + condition + ")"
                else:
                    # These parentheses could be of importance, check the leading
                    # "(" is not part of a smaller sub-statement:
                    if any(
                        level < 0 for _, level in self.find_and_match_braces(condition)
                    ):
                        continue

                    prefix = prefix[:-1]  # Remove parentheses
                    suffix = suffix[1:]

                    # Removing the () could cause a syntax error:
                    if not prefix.endswith(" "):
                        prefix += " "
                    if not suffix.startswith(" "):
                        suffix = " " + suffix

                self.add_correction(
                    (
                        "Parentheses around conditions are expected"
                        if self._parentheses
                        else "Parentheses around condition should be removed"
                    ),
                    i,
                )

                content[i] = prefix + condition + suffix

    @staticmethod
    def find_and_match_braces(
        text: str, brace_left: str = "(", brace_right: str = ")"
    ) -> Tuple[int, int]:
        """Step through braces in a string.

        Note that levels can step into negative.

        :return:    Tuple of (strpos, level), where strpos is the zero-index position of
                    the brace itself and level is the nested level it indicates
        """
        level = 0
        re_any_brace = re.compile(r"[" + brace_left + brace_right + "]")
        for match in re_any_brace.finditer(text):
            if match.group() == brace_left:
                level += 1  # Nest deeper
            else:
                level -= 1  # Nest back out
            yield match.start(), level



import re
from datetime import datetime
from logging import Logger
from pathlib import Path
from typing import Optional, Set

from git import GitCommandError, Repo

from ..common import Tool


class GitSetter:
    """Helper class that uses properties to expose parsed Git info.

    This is a separate class to control protected methods.

    The 'dirty' feature by Git is modified to ignore some changed files for the
    dirty-state.
    """

    _EMPTY = "[empty]"
    _DATETIME_FORMAT = "%d-%m-%Y %H:%M:%S"

    def __init__(
        self, repo: Repo, logger: Logger, tolerate_dirty: Optional[Set[Path]] = None
    ):
        self._repo: Repo = repo
        self._logger = logger
        self._tolerate_dirty = {p.absolute() for p in tolerate_dirty}
        self._is_empty: bool = False

        try:
            _ = self._repo.head.object
        except ValueError as err:
            self._logger.warning(f"Repository is probably empty: {err}")
            self._is_empty = True

    @property
    def empty(self) -> bool:
        return self._is_empty

    def keyword_replace(self, match) -> str:
        """Callback for regex replacement."""
        keyword: str = match.group(1)  # Skipping the "{{" and "}}"

        if keyword.startswith("GIT_"):
            # Get value through local property
            if self._is_empty:
                return self._EMPTY
            try:
                return getattr(self, keyword.lower())
            except GitCommandError as err:
                self._logger.warning(f"Error in keyword '{keyword}': {err}")
                return self._EMPTY

        if keyword.startswith("git "):
            # Run function instead
            commands_str = keyword[4:]  # Strip "git "
            commands = commands_str.split(" ")
            func = getattr(self._repo.git, commands[0])  # Retrieve function handle
            return func(*commands[1:])  # Call, passing in remaining words as arguments

        raise ValueError(f"Unrecognized class of placeholder: {keyword}")

    def _exempt_dirty(self) -> bool:
        """

        :return: `True` if all modified files are in the tolerate list
                 Will return `False` if files were only created/deleted
        """
        if not self._tolerate_dirty:
            return False  # No extra info

        repo_path = Path(self._repo.working_dir).absolute()

        exempt = False
        for item in self._repo.index.diff(None):
            if item.change_type != "M":
                continue  # Only for modified files

            diff_path = repo_path / Path(item.a_path)
            if diff_path not in self._tolerate_dirty:
                return False  # Modified file that we're not exempting, 'dirty' is real

            exempt = True

        return exempt

    @property
    def git_hash(self) -> str:
        return self._repo.head.object.hexsha

    @property
    def git_hash_short(self) -> str:
        return self.git_hash[:8]

    @property
    def git_date(self) -> str:
        return self._repo.head.object.committed_datetime.strftime(self._DATETIME_FORMAT)

    @property
    def git_now(self) -> str:
        """Technically not a Git command at all, but useful anyway."""
        return datetime.now().strftime(self._DATETIME_FORMAT)

    @property
    def git_tag(self) -> str:
        """Get the last most relevant tag to the current commit."""
        # Getting the most relevant tag through `self._repo` is not so straightforward
        # Easier to get it with `describe`
        return self._repo.git.describe("--tags", "--abbrev=0")

    @property
    def git_version(self) -> str:
        """Use `git_tag` but parse the three version digits."""
        tag = self.git_tag
        re_version = re.compile(r"\d+\.\d+\.+\d+")
        match = re_version.search(tag)
        if not match:
            return "0.0.0"

        return match.group()

    @property
    def git_branch(self) -> str:
        try:
            return self._repo.active_branch.name
        except TypeError as err:
            if "HEAD is a detached" not in str(err):
                raise  # Re-raise error again, can't handle this

            return self._EMPTY  # In detached head, current branch is not valid

    @property
    def git_description(self):
        return self._repo.git.describe("--tags", "--always")

    @property
    def git_description_dirty(self):
        txt: str = self._repo.git.describe("--tags", "--always", "--dirty")
        if self._exempt_dirty():
            txt = txt.replace("-dirty", "")
        return txt

    @property
    def git_dirty(self) -> str:
        return "1" if self._repo.is_dirty() and not self._exempt_dirty() else "0"


class GitInfo(Tool):
    """Class to insert Git version info into a template.

    We use template keys in the source file like ``{{key}}``.
    This is done to avoid conflicts with TwinCAT source files. e.g. ``<key>`` might
    collide with XML brackets in ``$key`` the dollar sign is a key for string constants.
    """

    LOGGER_NAME = "git_info"

    CONFIG_KEY = "git_info"

    PATH_VARIABLES = ["template", "tolerate_dirty"]

    def __init__(self, *args):
        super().__init__(*args)

        self._re_keyword = re.compile(r"{{([^}]+)}}")

    @classmethod
    def set_arguments(cls, parser):
        super().set_arguments(parser)

        parser.description = (
            "Create a new file with version info from Git from a template."
        )
        parser.epilog = "Example: ``tc_git_info Version.TcGVL.template``"
        parser.add_argument(
            "template",
            help="Template file to be used for newly created file",
        )
        parser.add_argument(
            "--output",
            help="File path for the new output file (default: template file with the "
            "last extension stripped)",
            default=None,
        )
        parser.add_argument(
            "--repo",
            help="Path to use for the Git repository (default: "
            "use the first repository up from the template file)",
            default=None,
        )
        parser.add_argument(
            "--tolerate-dirty",
            "-t",
            help="Paths to files that are allowed to be modified without showing the "
            "'dirty' flag",
            action="append",
            type=str,
            default=[],
        )
        return parser

    def run(self) -> int:
        """Produce an info file based on template.

        This is largely a copy of https://github.com/RobertoRoos/git-substitute.
        This DRY violation is accepted to prevent a code dependency.
        """

        template_path = Path(self.args.template)

        with template_path.open("r", newline="") as fh:
            content = fh.read()  # Preserve line endings

        self.logger.debug(f"Read file `{template_path.absolute()}`")

        repo_path = Path(self.args.repo) if self.args.repo else template_path.parent

        repo = Repo(repo_path, search_parent_directories=True)

        git_setter = GitSetter(
            repo,
            logger=self.logger,
            tolerate_dirty={Path(f) for f in self.args.tolerate_dirty},
        )

        # Perform placeholder substitution:
        content, keywords_used = self._re_keyword.subn(
            git_setter.keyword_replace, content
        )

        # Now fix escaped characters:
        content = content.replace(r"\{\{", "{{").replace(r"\}\}", "}}")

        self.logger.info(f"Applied {keywords_used} keyword(s) to template")

        if self.args.dry:
            print(content)
            return 0

        if self.args.output is None:
            if not template_path.suffix:
                self.logger.warning("Template file does not have a double extension")

            output_path = template_path.with_suffix("")
        else:
            output_path = Path(self.args.output)

        if keywords_used == 0:
            self.logger.error("Couldn't find any keywords to replace in template")
            return 1

        with open(output_path, "w", newline="") as fh:
            fh.write(content)

        self.logger.debug(f"Wrote to file `{output_path.absolute()}`")
        return 0



import logging
import sys
from abc import ABC, abstractmethod
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser, Namespace
from pathlib import Path
from typing import Any, Dict, List, Optional

from lxml import etree

from . import __version__

try:
    import tomllib
except ImportError:
    import tomli as tomllib  # Python 3.10 and before doesn't have tomllib yet

# Create type hinting shortcuts:
Element = etree._Element  # noqa
ElementTree = etree._ElementTree  # noqa


class Tool(ABC):
    """Tools base class.

    ``argparse`` is done in the constructor, CLI arguments should be passed there.
    """

    LOGGER_NAME: Optional[str] = None

    # Default value for file filter argument:
    FILTER_DEFAULT: List[str]

    CONFIG_KEY: Optional[str] = None

    PATH_VARIABLES: List[str] = []  # Names of options that are considered file paths

    def __init__(self, *args):
        """Pass e.g. ``sys.args[1:]`` (skipping the script part of the arguments).

        :param args: See :meth:`set_arguments`
        """
        parser = self.get_argument_parser()

        # All the fields were are going to have in the arguments output:
        fields = {
            action.dest for action in parser._actions if action.dest != "help"  # noqa
        }

        self.config_file: Optional[Path] = None

        config = self.make_config()
        if self.CONFIG_KEY:
            config = config.get(self.CONFIG_KEY, {})

        # Change the default values of the parser to the config results:
        for key, value in config.items():
            if key not in fields:
                raise ValueError(
                    f"Config field `{key}` is not recognized as a valid option"
                )
            if key in self.PATH_VARIABLES:
                value = self._make_path_from_config(value)
                # We should treat these paths as relative to the config file, ignoring
                # the current working directory

            parser.set_defaults(**{key: value})

            for action in parser._actions:
                if action.dest == key:
                    action.required = False
                    # Unfortunately this wouldn't disable required flags yet

        # Now parse CLI options on top of file configurations:
        self.args: Namespace = parser.parse_args(args)

        # Only after parsing all arguments we can establish a logger,
        self.logger = self.get_logger()

        if self.config_file:
            # Use property for config file path so we can still log it now:
            self.logger.debug(f"Loading from config: {self.config_file}")

    @classmethod
    def get_argument_parser(cls) -> ArgumentParser:
        parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
        cls.set_arguments(parser)
        return parser

    @classmethod
    def set_arguments(cls, parser):
        """Create application-specific arguments."""
        parser.add_argument(
            "--version", "-V", action="version", version="%(prog)s " + __version__
        )
        parser.add_argument(
            "--dry",
            help="Do not modify files on disk, only report changes to be made",
            action="store_true",
            default=False,
        )
        parser.add_argument(
            "--log-level",
            "-l",
            choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
            help="Set log level to change verbosity",
            default="INFO",
        )
        return

    def make_config(self) -> Dict[str, Any]:
        """Get configuration from possible files."""
        config = {}
        self.config_file = self._find_files_upwards(
            Path.cwd(), ["tctools.toml", "pyproject.toml"]
        )
        if not self.config_file:
            return config

        with open(self.config_file, "rb") as fh:
            data = tomllib.load(fh)
            if "tctools" in data:
                config = data["tctools"]

        return config

    @classmethod
    def _find_files_upwards(
        cls, directory: Path, filenames: List[str]
    ) -> Optional[Path]:
        """Find a file with a given name in the directory or it's parents.

        First hit on any of the filenames is returned.
        """
        for option in ([directory], directory.parents):
            for test_dir in option:
                for filename in filenames:
                    test_path = test_dir / filename
                    if test_path.is_file():
                        return test_path

        return None

    def _make_path_from_config(self, path: Any) -> Path:
        """Turn a relative path from a config file into a global path.

        Otherwise, return it as-is.

        :param path: Config value, can be (list of) `Path` or `str`
        """

        def _fix_path(p):
            p = Path(p)
            if not p.is_absolute() and self.config_file:
                p = self.config_file.parent / p
            return p

        if isinstance(path, list):
            path = [_fix_path(p) for p in path]
        else:
            path = _fix_path(path)

        return path

    @abstractmethod
    def run(self) -> int:
        """Main tool execution."""

    def get_logger(self):
        """Get logger for this class."""
        logging.basicConfig(stream=sys.stdout)
        logger = logging.getLogger(self.LOGGER_NAME or __name__)
        if "log_level" in self.args and self.args.log_level:
            level = logging.getLevelName(self.args.log_level)
            logger.setLevel(level)

        return logger


class TcTool(Tool, ABC):
    """Base class for tools sharing TwinCAT functionality."""

    PATH_VARIABLES = ["target"]

    def __init__(self, *args):
        super().__init__(*args)

        # Preserve `CDATA` XML flags:
        self.xml_parser = etree.XMLParser(strip_cdata=False)

        self.header_before: Optional[str] = None  # Header of the last XML path

        self.files_checked = 0  # Files read by parser
        self.files_to_alter = 0  # Files that seem to require changes
        self.files_resaved = 0  # Files actually re-saved to disk

    @classmethod
    def set_arguments(cls, parser):
        super().set_arguments(parser)

        parser.add_argument(
            "target",
            help="File(s) or folder(s) to target",
            nargs="+",
        )
        parser.add_argument(
            "--check",
            help="Do not modify files on disk, but give a non-zero exit code if there "
            "would be changes",
            action="store_true",
            default=False,
        )
        parser.add_argument(
            "-r",
            "--recursive",
            help="Also target folder (and their files) inside a target folder",
            action="store_true",
            default=False,
        )
        parser.add_argument(
            "--filter",
            help="Target files only with these patterns",
            nargs="+",
            default=cls.FILTER_DEFAULT,
        )

        return parser

    @staticmethod
    def get_xml_header(file: str) -> Optional[str]:
        """Get raw XML header as string."""
        with open(file, "r") as fh:
            # Search only the start of the path, otherwise give up
            for _ in range(100):
                line = fh.readline()
                if line.startswith("<?xml") and line.rstrip().endswith("?>"):
                    return line.strip()

        return None

    def get_xml_tree(self, path: str) -> ElementTree:
        """Get parsed XML path."""
        tree = etree.parse(path, self.xml_parser)

        self.header_before = self.get_xml_header(path)

        return tree

    def find_files(self) -> List[Path]:
        """Use argparse arguments to get a set of target files."""
        files = []
        if not self.args.target:
            return files

        targets = self.args.target
        if isinstance(targets, str) or isinstance(targets, Path):
            targets = [targets]

        for target in targets:
            path = Path(target).resolve()
            if path.is_file():
                files.append(path)
            elif path.is_dir():
                if self.args.filter:
                    for filt in self.args.filter:
                        if self.args.recursive:
                            filt = f"**/{filt}"
                        files += path.glob(filt)
            else:
                raise ValueError(f"Could not find path or folder: `{target}`")

        return files
